"""Asynchronous Python client for the WAQI API."""


class WAQIError(Exception):
    """Generic exception."""


class WAQIConnectionError(WAQIError):
    """WAQI connection exception."""


class WAQIUnknownCityError(WAQIError):
    """WAQI unknown city exception."""


class WAQIUnknownStationError(WAQIError):
    """WAQI unknown station exception."""


class WAQIAuthenticationError(WAQIError):
    """WAQI authentication exception."""



"""Fixtures for the aiowaqi package."""

from typing import AsyncGenerator

import aiohttp
import pytest

from aiowaqi import WAQIClient
from syrupy import SnapshotAssertion

from .syrupy import WAQISnapshotExtension


@pytest.fixture(name="snapshot")
def snapshot_assertion(snapshot: SnapshotAssertion) -> SnapshotAssertion:
    """Return snapshot assertion fixture with the WAQI extension."""
    return snapshot.use_extension(WAQISnapshotExtension)


@pytest.fixture(name="waqi_client")
async def client() -> AsyncGenerator[WAQIClient, None]:
    """Return a WAQI client."""
    async with aiohttp.ClientSession() as session, WAQIClient(
        session=session,
    ) as waqi_client:
        yield waqi_client


@pytest.fixture(name="authenticated_client")
async def authenticated_client(
    waqi_client: WAQIClient,
) -> WAQIClient:
    """Return an authenticated WAQI client."""
    waqi_client.authenticate("test")
    return waqi_client



"""Tests for the WAQI Library."""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

import aiohttp
from aiohttp.web_request import BaseRequest
from aresponses import Response, ResponsesMockServer
import pytest

from aiowaqi import (
    WAQIAirQuality,
    WAQIAuthenticationError,
    WAQIClient,
    WAQIConnectionError,
    WAQIError,
    WAQISearchResult,
    WAQIUnknownStationError,
)

from . import load_fixture

if TYPE_CHECKING:
    from syrupy import SnapshotAssertion

WAQI_URL = "api.waqi.info"


@pytest.mark.parametrize(
    "city",
    [
        "utrecht",
        "maarssen",
        "klundert",
        "olivias",
        "failing_klundert",
    ],
)
async def test_by_city(
    authenticated_client: WAQIClient,
    aresponses: ResponsesMockServer,
    city: str,
    snapshot: SnapshotAssertion,
) -> None:
    """Test retrieving air quality by city."""
    aresponses.add(
        WAQI_URL,
        f"/feed/{city}?token=test",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture(f"city_feed_{city}.json"),
        ),
        match_querystring=True,
    )
    response: WAQIAirQuality = await authenticated_client.get_by_city(city)
    assert response == snapshot


async def test_new_dominant_pol(
    authenticated_client: WAQIClient,
    aresponses: ResponsesMockServer,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test retrieving new dominant pol."""
    aresponses.add(
        WAQI_URL,
        "/feed/maarssen?token=test",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("city_feed_new_dominant_pol.json"),
        ),
        match_querystring=True,
    )
    await authenticated_client.get_by_city("maarssen")
    assert (
        "'h' is an unsupported value for <enum 'Pollutant'>,"
        " please report this at https://github.com/joostlek/python-waqi/issues"
        in caplog.text
    )


async def test_unknown_dominant_pol(
    aresponses: ResponsesMockServer,
    authenticated_client: WAQIClient,
) -> None:
    """Test retrieving unknown dominant pol."""
    aresponses.add(
        WAQI_URL,
        "/feed/maarssen?token=test",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("city_feed_unknown_dominant_pol.json"),
        ),
        match_querystring=True,
    )
    air_quality = await authenticated_client.get_by_city("maarssen")
    assert air_quality.dominant_pollutant is None


async def test_own_session(
    aresponses: ResponsesMockServer,
) -> None:
    """Test creating own session."""
    aresponses.add(
        WAQI_URL,
        "/feed/utrecht",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("city_feed_utrecht.json"),
        ),
    )
    async with WAQIClient() as waqi:
        assert waqi.session is None
        waqi.authenticate("test")
        await waqi.get_by_city("utrecht")
        assert waqi.session is not None


async def test_unexpected_server_response(
    authenticated_client: WAQIClient,
    aresponses: ResponsesMockServer,
) -> None:
    """Test handling unexpected response."""
    aresponses.add(
        WAQI_URL,
        "/feed/utrecht",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "plain/text"},
            text="Yes",
        ),
    )
    with pytest.raises(WAQIError):
        assert await authenticated_client.get_by_city("utrecht")


async def test_unknown_city(
    authenticated_client: WAQIClient,
    aresponses: ResponsesMockServer,
) -> None:
    """Test unknown city."""
    aresponses.add(
        WAQI_URL,
        "/feed/unknown",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("city_feed_unknown.json"),
        ),
    )
    with pytest.raises(WAQIError):
        assert await authenticated_client.get_by_city("unknown")


async def test_unauthenticated(
    authenticated_client: WAQIClient,
    aresponses: ResponsesMockServer,
) -> None:
    """Test unauthenticated."""
    aresponses.add(
        WAQI_URL,
        "/feed/utrecht",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("unauthenticated.json"),
        ),
    )
    with pytest.raises(WAQIAuthenticationError):
        assert await authenticated_client.get_by_city("utrecht")


async def test_timeout(aresponses: ResponsesMockServer) -> None:
    """Test request timeout."""

    # Faking a timeout by sleeping
    async def response_handler(_: BaseRequest) -> Response:
        """Response handler for this test."""
        await asyncio.sleep(2)
        return aresponses.Response(body="Goodmorning!")

    aresponses.add(
        WAQI_URL,
        "/feed/utrecht",
        "GET",
        response_handler,
    )
    async with aiohttp.ClientSession() as session:
        waqi = WAQIClient(session=session, request_timeout=1)
        waqi.authenticate("test")
        with pytest.raises(WAQIConnectionError):
            assert await waqi.get_by_city("utrecht")
        await waqi.close()


@pytest.mark.parametrize(
    "keyword",
    [
        "klundert",
        "failing_klundert",
        "unknown",
    ],
)
async def test_search(
    authenticated_client: WAQIClient,
    aresponses: ResponsesMockServer,
    keyword: str,
    snapshot: SnapshotAssertion,
) -> None:
    """Test searching stations."""
    aresponses.add(
        WAQI_URL,
        f"/search/?keyword={keyword}&token=test",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture(f"search_{keyword}.json"),
        ),
        match_querystring=True,
    )
    response: list[WAQISearchResult] = await authenticated_client.search(keyword)
    assert response == snapshot


@pytest.mark.parametrize(
    "name",
    [
        "klundert",
        "failing_klundert",
    ],
)
async def test_get_by_name(
    authenticated_client: WAQIClient,
    aresponses: ResponsesMockServer,
    name: str,
    snapshot: SnapshotAssertion,
) -> None:
    """Test getting stations by name."""
    aresponses.add(
        WAQI_URL,
        f"/feed/{name}?token=test",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture(f"name_feed_{name}.json"),
        ),
        match_querystring=True,
    )
    response = await authenticated_client.get_by_name(name)
    assert response == snapshot


async def test_get_unknown_by_name(
    authenticated_client: WAQIClient,
    aresponses: ResponsesMockServer,
) -> None:
    """Test getting unknown station by name."""
    aresponses.add(
        WAQI_URL,
        "/feed/unknown",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("name_feed_unknown.json"),
        ),
    )
    with pytest.raises(WAQIUnknownStationError):
        await authenticated_client.get_by_name("unknown")


@pytest.mark.parametrize(
    "station_number",
    [
        6337,
        372382,
        10142,
        10002,
    ],
)
async def test_get_by_station_number(
    authenticated_client: WAQIClient,
    aresponses: ResponsesMockServer,
    station_number: int,
    snapshot: SnapshotAssertion,
) -> None:
    """Test getting stations by station_number."""
    aresponses.add(
        WAQI_URL,
        f"/feed/@{station_number}?token=test",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture(f"station_number_feed_{station_number}.json"),
        ),
        match_querystring=True,
    )
    response = await authenticated_client.get_by_station_number(station_number)
    assert response == snapshot


@pytest.mark.parametrize(
    "identifier",
    [
        "unknown",
        "@123946",
        "A10142",
    ],
)
async def test_get_unknown_by_station_number(
    authenticated_client: WAQIClient,
    aresponses: ResponsesMockServer,
    identifier: str,
) -> None:
    """Test getting unknown station by station_number."""
    aresponses.add(
        WAQI_URL,
        "/feed/@0",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture(f"station_number_feed_{identifier}.json"),
        ),
    )
    with pytest.raises(WAQIUnknownStationError):
        await authenticated_client.get_by_station_number(0)


async def test_get_by_coordinates(
    authenticated_client: WAQIClient,
    aresponses: ResponsesMockServer,
    snapshot: SnapshotAssertion,
) -> None:
    """Test getting measuring station via coordinates."""
    aresponses.add(
        WAQI_URL,
        "/feed/geo:52.105031;5.124464",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("coordinates.json"),
        ),
    )
    response = await authenticated_client.get_by_coordinates(52.105031, 5.124464)
    assert response == snapshot


async def test_get_by_ip(
    authenticated_client: WAQIClient,
    aresponses: ResponsesMockServer,
    snapshot: SnapshotAssertion,
) -> None:
    """Test getting measuring station via ip."""
    aresponses.add(
        WAQI_URL,
        "/feed/here",
        "GET",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/json"},
            text=load_fixture("here.json"),
        ),
    )
    response = await authenticated_client.get_by_ip()
    assert response == snapshot



"""Asynchronous Python client for the WAQI API."""

from __future__ import annotations

from typing import Any, TypeVar

from .const import LOGGER

_EnumT = TypeVar("_EnumT")


def to_nullable_enum(
    enum_class: type[_EnumT],
    value: Any,
) -> _EnumT | None:
    """Convert a value to an enum and log if it doesn't exist."""
    try:
        return enum_class(value)  # type: ignore[call-arg]
    except ValueError:
        LOGGER.warning(
            "%r is an unsupported value for %s, please report this at https://github.com/joostlek/python-waqi/issues",
            value,
            str(enum_class),
        )
        return None



"""Asynchronous Python client for the WAQI API."""

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from importlib import metadata
from typing import TYPE_CHECKING, Any, cast

from aiohttp import ClientSession
from aiohttp.hdrs import METH_GET
from yarl import URL

from .exceptions import (
    WAQIAuthenticationError,
    WAQIConnectionError,
    WAQIError,
    WAQIUnknownCityError,
    WAQIUnknownStationError,
)
from .models import WAQIAirQuality, WAQISearchResult

if TYPE_CHECKING:
    from typing_extensions import Self

VERSION = metadata.version(__package__)


@dataclass
class WAQIClient:
    """Main class for handling connections with WAQI."""

    session: ClientSession | None = None
    request_timeout: int = 10
    api_host: str = "api.waqi.info"
    _token: str | None = None
    _close_session: bool = False

    def authenticate(self, token: str) -> None:
        """Authenticate the user with a token."""
        self._token = token

    async def _request(
        self,
        uri: str,
        *,
        data: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Handle a request to WAQI.

        A generic method for sending/handling HTTP requests done against
        WAQI.

        Args:
        ----
            uri: the path to call.
            data: the query parameters to add.

        Returns:
        -------
            A Python dictionary (JSON decoded) with the response from
            the API.

        Raises:
        ------
            WAQIConnectionError: An error occurred while communicating with
                the WAQI API.
            WAQIError: Received an unexpected response from the WAQI API.
            WAQIAuthenticationError: Used token is invalid.

        """
        url = URL.build(
            scheme="https",
            host=self.api_host,
            port=443,
        ).joinpath(uri)

        headers = {
            "User-Agent": f"WAQIAsync/{VERSION}",
            "Accept": "application/json, text/plain, */*",
        }

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        if data is None:  # pragma: no cover
            data = {}
        data["token"] = self._token
        try:
            async with asyncio.timeout(self.request_timeout):
                response = await self.session.request(
                    METH_GET,
                    url.with_query(data),
                    headers=headers,
                )
        except asyncio.TimeoutError as exception:
            msg = "Timeout occurred while connecting to the WAQI API"
            raise WAQIConnectionError(msg) from exception

        content_type = response.headers.get("Content-Type", "")

        if "application/json" not in content_type:
            text = await response.text()
            msg = "Unexpected response from the WAQI API"
            raise WAQIError(
                msg,
                {"Content-Type": content_type, "response": text},
            )

        response_data = cast(dict[str, Any], await response.json())
        if (
            response_data["status"] == "error"
            and response_data["data"] == "Invalid key"
        ):
            raise WAQIAuthenticationError
        return response_data

    async def get_by_city(self, city: str) -> WAQIAirQuality:
        """Get air quality information for a given city."""
        response = await self._request(f"feed/{city}")
        if response["status"] == "error" and response["data"] == "Unknown station":
            msg = f"Could not find city {city}"
            raise WAQIUnknownCityError(msg)
        return WAQIAirQuality.from_dict(response["data"])

    async def get_by_name(self, name: str) -> WAQIAirQuality:
        """Get air quality measuring station by name."""
        response = await self._request(f"feed/{name}")
        data = response["data"]
        if (
            response["status"] == "error"
            and data in ("Unknown station", "no such station")
        ) or (
            "status" in data
            and data["status"] == "error"
            and data["msg"] == "Unknown ID"
        ):
            msg = f"Could not find station {name}"
            raise WAQIUnknownStationError(msg)
        return WAQIAirQuality.from_dict(data)

    async def get_by_station_number(self, station_number: int) -> WAQIAirQuality:
        """Get air quality measuring station by station number."""
        return await self.get_by_name(f"@{station_number}")

    async def get_by_coordinates(
        self,
        latitude: float,
        longitude: float,
    ) -> WAQIAirQuality:
        """Get nearest air quality measuring station by coordinates."""
        response = await self._request(f"feed/geo:{latitude};{longitude}")
        return WAQIAirQuality.from_dict(response["data"])

    async def get_by_ip(
        self,
    ) -> WAQIAirQuality:
        """Get the nearest air quality measuring station according to WAQI."""
        return await self.get_by_name("here")

    async def search(self, keyword: str) -> list[WAQISearchResult]:
        """Search for stations with a keyword."""
        response = await self._request("search/", data={"keyword": keyword})
        return [WAQISearchResult.from_dict(station) for station in response["data"]]

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The WAQIClient object.

        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.

        """
        await self.close()



"""Asynchronous Python client for the WAQI API."""

from __future__ import annotations

from contextlib import suppress
from dataclasses import dataclass
from datetime import datetime
from enum import StrEnum
from typing import Any, Self

from aiowaqi.util import to_nullable_enum


class Pollutant(StrEnum):
    """Enum of pollutants."""

    CARBON_MONOXIDE = "co"
    NITROGEN_DIOXIDE = "no2"
    OZONE = "o3"
    SULFUR_DIOXIDE = "so2"
    PM10 = "pm10"
    PM25 = "pm25"
    NEPHELOMETRY = "neph"


@dataclass(slots=True)
class Attribution:
    """Represents an attribution."""

    url: str
    name: str
    logo: str | None

    @classmethod
    def from_dict(cls, attribution: dict[str, Any]) -> Self:
        """Initialize from a dict."""
        return cls(
            url=attribution["url"],
            name=attribution["name"].strip(),
            logo=attribution.get("logo"),
        )


@dataclass(slots=True)
class Coordinates:
    """Represents coordinates."""

    latitude: float
    longitude: float


@dataclass(slots=True)
class Location:
    """Represents a location object."""

    external_url: str
    name: str
    coordinates: Coordinates

    @classmethod
    def from_dict(cls, location: dict[str, Any]) -> Self:
        """Initialize from a dict."""
        return cls(
            external_url=location["url"],
            name=location["name"],
            coordinates=Coordinates(
                latitude=location["geo"][0],
                longitude=location["geo"][1],
            ),
        )


@dataclass(slots=True)
class City(Location):
    """Represents a city object."""

    location: str | None

    @classmethod
    def from_dict(cls, location: dict[str, Any]) -> Self:
        """Initialize from a dict."""
        loc = location["location"]
        if not loc:
            loc = None
        return cls(
            external_url=location["url"],
            name=location["name"],
            coordinates=Coordinates(
                latitude=location["geo"][0],
                longitude=location["geo"][1],
            ),
            location=loc,
        )


@dataclass(slots=True)
class Station(Location):
    """Represents a station object."""


@dataclass(slots=True)
class WAQIExtendedAirQuality:
    """Represents extended air quality data."""

    carbon_monoxide: float | None
    humidity: float | None
    nephelometry: float | None
    nitrogen_dioxide: float | None
    ozone: float | None
    pressure: float | None
    sulfur_dioxide: float | None
    pm10: float | None
    pm25: float | None
    temperature: float | None

    @classmethod
    def from_dict(cls, air_quality: dict[str, Any]) -> Self:
        """Initialize from a dict."""
        return cls(
            carbon_monoxide=air_quality.get("co", {}).get("v"),
            humidity=air_quality.get("h", {}).get("v"),
            nephelometry=air_quality.get("neph", {}).get("v"),
            nitrogen_dioxide=air_quality.get("no2", {}).get("v"),
            ozone=air_quality.get("o3", {}).get("v"),
            pressure=air_quality.get("p", {}).get("v"),
            sulfur_dioxide=air_quality.get("so2", {}).get("v"),
            pm10=air_quality.get("pm10", {}).get("v"),
            pm25=air_quality.get("pm25", {}).get("v"),
            temperature=air_quality.get("t", {}).get("v"),
        )


@dataclass(slots=True)
class WAQIAirQuality:
    """Represents the air quality data from WAQI."""

    air_quality_index: int | None
    station_id: int
    attributions: list[Attribution]
    city: City
    extended_air_quality: WAQIExtendedAirQuality
    dominant_pollutant: Pollutant | None
    measured_at: datetime | None

    @classmethod
    def from_dict(cls, air_quality: dict[str, Any]) -> Self:
        """Initialize from a dict."""
        aqi = None
        with suppress(ValueError):
            aqi = int(air_quality["aqi"])

        dominant_pollutant = air_quality["dominentpol"]
        if dominant_pollutant == "":
            dominant_pollutant = None
        else:
            dominant_pollutant = to_nullable_enum(Pollutant, air_quality["dominentpol"])

        measured_at: datetime | None = None
        if "iso" in air_quality["time"]:
            measured_at = datetime.fromisoformat(air_quality["time"]["iso"])

        return cls(
            air_quality_index=aqi,
            station_id=air_quality["idx"],
            attributions=[
                Attribution.from_dict(attribution)
                for attribution in air_quality["attributions"]
            ],
            city=City.from_dict(air_quality["city"]),
            extended_air_quality=WAQIExtendedAirQuality.from_dict(air_quality["iaqi"]),
            dominant_pollutant=dominant_pollutant,
            measured_at=measured_at,
        )


@dataclass(slots=True)
class WAQISearchResult:
    """Represents a search result from the WAQI api."""

    air_quality_index: int | None
    station_id: int
    station: Station

    @classmethod
    def from_dict(cls, result: dict[str, Any]) -> Self:
        """Initialize from a dict."""
        aqi = None
        with suppress(ValueError):
            aqi = int(result["aqi"])

        return cls(
            air_quality_index=aqi,
            station_id=result["uid"],
            station=Station.from_dict(result["station"]),
        )



"""Asynchronous Python client for the WAQI API."""

from .exceptions import (
    WAQIAuthenticationError,
    WAQIConnectionError,
    WAQIError,
    WAQIUnknownCityError,
    WAQIUnknownStationError,
)
from .models import (
    Attribution,
    City,
    Coordinates,
    Location,
    WAQIAirQuality,
    WAQIExtendedAirQuality,
    WAQISearchResult,
)
from .waqi import WAQIClient

__all__ = [
    "WAQIError",
    "WAQIUnknownStationError",
    "WAQIUnknownCityError",
    "Location",
    "WAQISearchResult",
    "WAQIAuthenticationError",
    "WAQIClient",
    "WAQIConnectionError",
    "Attribution",
    "City",
    "Coordinates",
    "WAQIExtendedAirQuality",
    "WAQIAirQuality",
]
